#include "LiquidCrystal_I2C.h"
#include <unistd.h>

LiquidCrystal_I2C::LiquidCrystal_I2C (PropWare::I2CMaster *i2cbus,
                                      uint8_t lcd_addr,
                                      uint8_t lcd_cols,
                                      uint8_t lcd_rows,
                                      uint8_t charsize) {
    this->_addr         = lcd_addr;
    this->_cols         = lcd_cols;
    this->_rows         = lcd_rows;
    this->_charsize     = charsize;
    this->_backlightval = LCD_BACKLIGHT;
    this->_i2cbus       = i2cbus;
}

int LiquidCrystal_I2C::begin () {
    int err;

    this->_displayfunction = LCD_4BITMODE | LCD_1LINE | LCD_5x8DOTS;
    if (this->_rows > 1) this->_displayfunction |= LCD_2LINE;
    if ((this->_charsize != 0) && (this->_rows == 1)) this->_displayfunction |= LCD_5x10DOTS;

    LiquidCrystal_I2C::delay(50);


    if (!this->_i2cbus->ping(this->_addr << 1))
        return 2;
    check_errors(expanderWrite(this->_backlightval));
    LiquidCrystal_I2C::delay(1000);

    check_errors(write4bits(0x03 << 4));
    LiquidCrystal_I2C::delayMicro(4500);

    check_errors(write4bits(0x03 << 4));
    LiquidCrystal_I2C::delayMicro(4500);

    check_errors(write4bits(0x03 << 4));
    LiquidCrystal_I2C::delayMicro(150);

    check_errors(write4bits(0x02 << 4));

    check_errors(this->command(LCD_FUNCTIONSET | this->_displayfunction));

    this->_displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;
    check_errors(this->display());

    check_errors(this->clear());

    this->_displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;

    check_errors(this->command(LCD_ENTRYMODESET | this->_displaymode));

    check_errors(this->home());

    return 0;
}

int LiquidCrystal_I2C::clear () {
    int err;
    check_errors(this->command(LCD_CLEARDISPLAY));
    LiquidCrystal_I2C::delayMicro(2000);
    return 0;
}

int LiquidCrystal_I2C::home () {
    int err;
    check_errors(this->command(LCD_RETURNHOME));
    LiquidCrystal_I2C::delayMicro(2000);
    return 0;
}

int LiquidCrystal_I2C::noDisplay () {
    this->_displaycontrol &= ~LCD_DISPLAYON;
    return this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

int LiquidCrystal_I2C::display () {
    this->_displaycontrol |= LCD_DISPLAYON;
    return this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

int LiquidCrystal_I2C::noBlink () {
    this->_displaycontrol &= ~LCD_BLINKON;
    return this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

int LiquidCrystal_I2C::blink () {
    this->_displaycontrol |= LCD_BLINKON;
    return this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

int LiquidCrystal_I2C::noCursor () {
    this->_displaycontrol &= ~LCD_CURSORON;
    return this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

int LiquidCrystal_I2C::cursor () {
    this->_displaycontrol |= LCD_CURSORON;
    return this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

int LiquidCrystal_I2C::scrollDisplayLeft () {
    return this->command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVELEFT);
}

int LiquidCrystal_I2C::scrollDisplayRight () {
    return this->command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVERIGHT);
}

int LiquidCrystal_I2C::leftToRight () {
    this->_displaymode |= LCD_ENTRYLEFT;
    return this->command(LCD_ENTRYMODESET | this->_displaymode);
}

int LiquidCrystal_I2C::rightToLeft () {
    this->_displaymode &= ~LCD_ENTRYLEFT;
    return this->command(LCD_ENTRYMODESET | this->_displaymode);
}

int LiquidCrystal_I2C::noBacklight () {
    this->_backlightval = LCD_NOBACKLIGHT;
    return this->expanderWrite(0);
}

int LiquidCrystal_I2C::backlight () {
    this->_backlightval = LCD_BACKLIGHT;
    return this->expanderWrite(0);
}

bool LiquidCrystal_I2C::getBacklight () {
    return this->_backlightval == LCD_BACKLIGHT;
}

int LiquidCrystal_I2C::autoscroll () {
    this->_displaymode |= LCD_ENTRYSHIFTINCREMENT;
    return this->command(LCD_ENTRYMODESET | this->_displaymode);
}

int LiquidCrystal_I2C::noAutoscroll () {
    this->_displaymode &= ~LCD_ENTRYSHIFTINCREMENT;
    return this->command(LCD_ENTRYMODESET | this->_displaymode);
}

int LiquidCrystal_I2C::createChar (uint8_t location, uint8_t charmap[]) {
    int err;
    location &= 0x7; // we only have 8 locations 0-7
    check_errors(this->command(LCD_SETCGRAMADDR | (location << 3)));
    for (int i = 0; i < 8; i++) check_errors(this->write(charmap[i]));
    return 0;
}

int LiquidCrystal_I2C::setCursor (uint8_t col, uint8_t row) {
    int row_offsets[] = {0x00, 0x40, 0x14, 0x54};
    if (row > this->_rows) row = this->_rows - 1;    // we count rows starting w/0
    return this->command(LCD_SETDDRAMADDR | (col + row_offsets[row]));
}

int LiquidCrystal_I2C::command (uint8_t value) {
    return this->send(value, 0);
}

int LiquidCrystal_I2C::write (uint8_t value) {
    return this->send(value, Rs);
}

int LiquidCrystal_I2C::blink_on () {
    return this->blink();
}

int LiquidCrystal_I2C::blink_off () {
    return this->noBlink();
}

int LiquidCrystal_I2C::cursor_on () {
    return this->cursor();
}

int LiquidCrystal_I2C::cursor_off () {
    return this->noCursor();
}

int LiquidCrystal_I2C::setBacklight (bool new_val) {
    if (new_val) return this->backlight();        // turn backlight on
    else return this->noBacklight();        // turn backlight off

    //new_val ? this->backlight() : this->noBacklight();
}

int LiquidCrystal_I2C::load_custom_character (uint8_t char_num, uint8_t *rows) {
    return this->createChar(char_num, rows);
}

int LiquidCrystal_I2C::printstr (const char c[]) {
    int         err;
    for (size_t i = 0; i < strlen(c); i++) check_errors(this->write(c[i]));
    return 0;
}

//------------------------------
int LiquidCrystal_I2C::send (uint8_t value, uint8_t mode) {
    int     err;
    uint8_t highnib = value & 0xf0;
    uint8_t lownib  = (value << 4) & 0xf0;
    check_errors(this->write4bits((highnib) | mode));
    check_errors(this->write4bits((lownib) | mode));
    return 0;
}

int LiquidCrystal_I2C::write4bits (uint8_t value) {
    int err;
    check_errors(this->expanderWrite(value));
    check_errors(this->pulseEnable(value));
    return 0;
}

int LiquidCrystal_I2C::expanderWrite (uint8_t _data) {
    if (this->_i2cbus->put(this->_addr << 1, (int) (_data) | this->_backlightval))
        return 0;
    else
        return 2;
}

int LiquidCrystal_I2C::pulseEnable (uint8_t _data) {
    int err;
    check_errors(this->expanderWrite(_data | En));
    LiquidCrystal_I2C::delayMicro(1);

    check_errors(this->expanderWrite(_data & ~En));
    LiquidCrystal_I2C::delayMicro(50);
    return 0;
}
//------------------------------
