#include <stdio.h>            // Recommended over iostream for saving space
//#include <propeller.h>        // Propeller-specific functions
//#include <simpletools.h>

#include <PropWare/PropWare.h>
#include <PropWare/gpio/simpleport.h>
#include <PropWare/serial/spi/spi.h>
#include <PropWare/serial/i2c/i2cmaster.h>

//extern _Driver _SimpleSerialDriver;
//extern _Driver _FileDriver;
//extern _Driver _NullDriver;

/* Driver list */
/*_Driver *_driverlist[] = {
	&_NullDriver,
	//&_SimpleSerialDriver,
	&_FileDriver,
	NULL
};*/

#include "LiquidCrystal_I2C.h"
#include "LTC6903.h"
#include "SN76489.h"
#include "YM2612.h"

#define TestSong 1

#if TestSong
#include "song.h"
#endif

//void mount(void);

//YM2612::sleep37_5(x)
#define wait_samples(x) usleep(x)

//#define SDA 29
//#define SCL 28
#define LCD_ADDR 0x27

//#define SCLK 26
//#define SDI 25
#define YM_CLK_CS PropWare::Pin::Mask::P23 //23
#define SN_CLK_CS PropWare::Pin::Mask::P22 //22

#define YM_CS PropWare::Pin::Mask::P14 //14
#define YM_WR PropWare::Pin::Mask::P11 //11
#define YM_RD PropWare::Pin::Mask::P10 //10
#define YM_A0 PropWare::Pin::Mask::P8 //8
#define YM_A1 PropWare::Pin::Mask::P9 //9
#define YM_IC PropWare::Pin::Mask::P12 //12
#define YM_RDY PropWare::Pin::Mask::P7 //7 //Ready bit on the data bus.

#define SN_CS PropWare::Pin::Mask::P15 //15

/*#define YM_CS 14
#define YM_WR 11
#define YM_RD 10
#define YM_A0 8
#define YM_A1 9
#define YM_IC 12

#define SN_CS 15*/

//#define SD_DO 27
//#define SD_CLK 26
//#define SD_DI 25
//#define SD_CS 24

#define MISO PropWare::Pin::Mask::P27 //27
#define SCLK PropWare::Pin::Mask::P26 //26
#define MOSI PropWare::Pin::Mask::P25 //25

#define SD_CS PropWare::Pin::Mask::P24 //24

int debugLCD(LiquidCrystal_I2C &lcd, const char msg[]) {
    int err;
	//printf("%s\n", msg);

	check_errors(lcd.clear());
	check_errors(lcd.setCursor(0, 0));
	check_errors(lcd.printstr(msg));

	return 0;
}

void SendData(YM2612 &ym2612, uint32_t data);

int run() {
    int err;

    PropWare::SPI spibus_osc(MISO, MISO, SCLK);

    LTC6903 ym_clk(&spibus_osc, YM_CLK_CS);
    LTC6903 sn_clk(&spibus_osc, SN_CLK_CS);

    PropWare::I2CMaster i2c;
    LiquidCrystal_I2C   lcd(&i2c, LCD_ADDR, 16, 2);

    PropWare::SimplePort bus(PropWare::Pin::Mask::P0, 8, PropWare::Port::Dir::OUT);

    YM2612 ym2612(&bus, YM_CS, YM_RD, YM_WR, YM_IC, YM_A0, YM_A1, YM_RDY);
    SN76489 sn76489(&bus, SN_CS);

    pwOut << "Hello!\n";

    //i2c_open(&i2c, SCL, SDA, 0);

    int i = 0;
    check_errors(lcd.begin());
    pwOut << "Running " << i++ << '\n';
    check_errors(lcd.backlight());
    pwOut << "Running " << i++ << '\n';
    check_errors(lcd.setCursor(0, 0));
    pwOut << "Running " << i++ << '\n';
    check_errors(lcd.printstr("Hello, world!"));
    pwOut << "Running " << i++ << '\n';
    check_errors(lcd.setCursor(0, 1));
    pwOut << "Running " << i++ << '\n';
    check_errors(lcd.printstr("Hello propeller!"));
    pwOut << "Running " << i++ << '\n';

    sleep(1);

    /*if (sd_mount(SD_DO, SD_CLK, SD_DI, SD_CS)) {
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.printstr("Error mounting");
        lcd.setCursor(0, 1);
        lcd.printstr("SD Card.");
        return 1;
    }*/

    spibus_osc.set_bit_mode(PropWare::SPI::BitMode::MSB_FIRST);
    spibus_osc.set_mode(PropWare::SPI::Mode::MODE_1);

    check_errors(debugLCD(lcd, "Freq."));
    //OCT: 12
    //DAC: 912
    //ym_clk.SetFrequency(7670453); //PAL 7600489 //NTSC 7670453
    ym_clk.SetManual(12, 912);
    //OCT: 11
    //DAC: 831
    //sn_clk.SetFrequency(3579545); //PAL 3546894 //NTSC 3579545
    sn_clk.SetManual(11, 831);

    sleep(1);


#if TestSong

    check_errors(debugLCD(lcd, "Res."));
    ym2612.Reset();
    sn76489.Reset();

    while (true) {
        check_errors(debugLCD(lcd, "Playing."));

        for (uint32_t i = 0; i < songlength; i++) SendData(ym2612, song[i]);

        check_errors(debugLCD(lcd, "Done!"));

        sleep(3);

        check_errors(debugLCD(lcd, "Res."));
        ym2612.Reset();
        sn76489.Reset();

        sleep(3);
    }

#else

    debugLCD(lcd, "Res.");
	ym2612.Reset();
	sn76489.Reset();

	debugLCD(lcd, "LFO Off.");
	ym2612.SendPacked(0x2200); //lfo off

	debugLCD(lcd, "CH 3 mode normal");
	ym2612.SendPacked(0x2700); //channel 3 mode normal

	debugLCD(lcd, "All Channels Off");
	for (int i = 1; i < 7; i++) ym2612.Send(0x28, i); //all channels off

	//DAC off
	debugLCD(lcd, "DAC off.");
	ym2612.SendPacked(0x2B00);
	ym2612.SendPacked(0x3071);
	ym2612.SendPacked(0x340D);
	ym2612.SendPacked(0x3833);

	//DT1/MUL
	debugLCD(lcd, "DT1/MUL");
	//ym2612.SendPacked(0x3071);
	//ym2612.SendPacked(0x340D);
	//ym2612.SendPacked(0x3833);
	ym2612.SendPacked(0x3C01);
	ym2612.SendPacked(0x4023);
	ym2612.SendPacked(0x442D);
	ym2612.SendPacked(0x4826);

	//Total Level
	debugLCD(lcd, "Total Level");
	//ym2612.SendPacked(0x4023);
	//ym2612.SendPacked(0x442D);
	//ym2612.SendPacked(0x4826);
	ym2612.SendPacked(0x4C00);
	ym2612.SendPacked(0x505F);
	ym2612.SendPacked(0x5499);
	ym2612.SendPacked(0x585F);

	//RS/AR
	debugLCD(lcd, "RS/AR");
	//ym2612.SendPacked(0x505F);
	//ym2612.SendPacked(0x5499);
	//ym2612.SendPacked(0x585F);
	ym2612.SendPacked(0x5C94);
	ym2612.SendPacked(0x6005);
	ym2612.SendPacked(0x6405);
	ym2612.SendPacked(0x6805);

	//AM/D1R
	debugLCD(lcd, "AM/D1R");
	//ym2612.SendPacked(0x6005);
	//ym2612.SendPacked(0x6405);
	//xym2612.SendPacked(0x6805);
	ym2612.SendPacked(0x6C07);
	ym2612.SendPacked(0x7002);
	ym2612.SendPacked(0x7402);
	ym2612.SendPacked(0x7802);

	//D2R
	debugLCD(lcd, "D2R");
	//ym2612.SendPacked(0x7002);
	//ym2612.SendPacked(0x7402);
	//ym2612.SendPacked(0x7802);
	ym2612.SendPacked(0x7C02);
	ym2612.SendPacked(0x8011);
	ym2612.SendPacked(0x8411);
	ym2612.SendPacked(0x8811);

	//D1L/RR
	debugLCD(lcd, "D1L/RR");
	//ym2612.SendPacked(0x8011);
	//ym2612.SendPacked(0x8411);
	//ym2612.SendPacked(0x8811);
	ym2612.SendPacked(0x8CA6);
	ym2612.SendPacked(0x9000);
	ym2612.SendPacked(0x9400);
	ym2612.SendPacked(0x9800);

	//Proprietary
	debugLCD(lcd, "Proprietary");
	//ym2612.SendPacked(0x9000);
	//ym2612.SendPacked(0x9400);
	//ym2612.SendPacked(0x9800);
	ym2612.SendPacked(0x9C00);

	debugLCD(lcd, "FB/Alg");
	ym2612.SendPacked(0xB032); //Feedback/algorithm

	debugLCD(lcd, "Speakers on");
	ym2612.SendPacked(0xB4C0); //Both Speakers on

	debugLCD(lcd, "Key off");
	ym2612.SendPacked(0x2800); //Key off

	//Set frequency
	debugLCD(lcd, "Set Freq");
	ym2612.SendPacked(0xA422);
	ym2612.SendPacked(0xA069);

	while (1) {
		debugLCD(lcd, "Key On");
		ym2612.SendPacked(0x28F0); //key on

		sleep(1);

		debugLCD(lcd, "Key Off 2");
		ym2612.SendPacked(0x2800); //key off

		sleep(1);
	}

#endif

    return 0;
}

int main() {
    const int err = run();
    if (err) {
        pwOut << "Oh no! We hit an error: " << err << '\n';
    }
    return err;
}

void SendData(YM2612 &ym2612, uint32_t data) {
	ym2612.SendPacked((uint16_t)(data & 0xFFFF));
	uint16_t delay = (uint16_t)((data & 0xFFFF0000) >> 16);
	if (delay != 0) wait_samples(delay);
}

/*void mount(void) {
	extern int dfs_mountflag;

	_SD_Params sdspi;
	sdspi.AttachmentType = _SDA_SingleSPI;
	sdspi.pins.SingleSPI.CLK = SD_CLK;
	sdspi.pins.SingleSPI.MISO = SD_DO;
	sdspi.pins.SingleSPI.MOSI = SD_DI;
	sdspi.pins.SingleSPI.CS = SD_CS;


	if (!dfs_mountflag)
		dfs_mount(&sdspi);
}*/
